#!/usr/bin/env python

import pyinotify
import os
from .packagefinder import PackageFinder


class ModuleWatcher(pyinotify.ProcessEvent):
    """
    Automatically reload any modules or packages as they change
    """

    def __init__(self, script_path, callback=None):
        "El constructor"

        self.wm = pyinotify.WatchManager()
        self.notifier = None

        pf = PackageFinder()
        self._callback = callback
        self._files = set(pf.get_files_list(script_path))
        self._directories = set()
        for filename in self._files:
            if not os.path.exists(filename):
                continue
            dirname = os.path.dirname(filename.rstrip('/'))
            self._directories.add(dirname)

        for dirname in self._directories:
            self.wm.add_watch(dirname, pyinotify.ALL_EVENTS)

    def start_watching(self):
        "Start the pyinotify watch thread"

        if self.notifier is None:
            self.notifier = pyinotify.ThreadedNotifier(self.wm, self)
        self.notifier.start()

    def stop_watching(self):
        "Stop the pyinotify watch thread"

        if self.notifier is not None:
            self.notifier.stop()

    def process_IN_MODIFY(self, event):
        "A file of interest has changed"

        # Is it a file I know about?
        if event.pathname not in self._files:
            return
        if self._callback is not None:
            getattr(self, '_callback')(event.pathname)



import os
from types import ModuleType
from importlib import import_module


# TODO make this function work with any python
# currently working wity python >= 3.5
def import_from_file(path):
    import importlib.util
    spec = importlib.util.spec_from_file_location('__not_main__', path)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    return foo


class PackageFinder(object):

    def _find_required_modules(self, module, processed_modules=None):
        if module.__name__ in ('__builtin__', '__future__'):
            return

        if processed_modules is None:
            processed_modules = set()

        if module in processed_modules:
            return processed_modules

        processed_modules.add(module)
        names = dir(module)
        for name in names:
            if name[0] == '_':
                continue
            obj = getattr(module, name)
            if isinstance(obj, ModuleType):
                self._find_required_modules(obj, processed_modules)
            elif hasattr(obj, '__module__') and obj.__module__ is not None:
                # in case of 'import somename from somemodule'
                # somename will be in globals and somemodule not
                # to get somemodule we need to get somename.__module__
                try:
                    obj = import_module(obj.__module__)
                    self._find_required_modules(obj, processed_modules)
                except ImportError:
                    pass

        return processed_modules

    def get_files_list(self, script_path):
        script_path = os.path.abspath(script_path)
        module = import_from_file(script_path)
        modules = self._find_required_modules(module)
        files = set()
        for module in modules:
            if not hasattr(module, '__file__') or module.__file__ == '':
                continue
            files.add(os.path.abspath(module.__file__))
        return list(files)



# coding: utf-8
from lazy import lazy
from math import sin, cos, pi, sqrt, atan2


class Vector(object):

    def __init__(self, x=None, y=None, z=None):
        if isinstance(x, list) and len(x) == 3 and y is None and z is None:
            x, y, z = x[0], x[1], x[2]

        self.x = x or 0
        self.y = y or 0
        self.z = z or 0

    @classmethod
    def com_for_children(cls, children):
        result = Vector()
        for child in children:
            result += child.com
        if children:
            result /= len(children)
        return result

    @classmethod
    def com_for_linear_extrude(cls, children, *args, **kwargs):
        result = cls.com_for_children(children)
        height = kwargs.get('h', args[0])
        return Vector(
            result.x,
            result.y,
            result.z + height / 2
        )

    @classmethod
    def com_for_cube(self, children, *args, **kwargs):
        x, y, z = args[0]
        return Vector(x / 2, y / 2, z / 2)

    @classmethod
    def com_for_cylinder(cls, children, *args, **kwargs):
        return Vector(0, 0, kwargs['h'] / 2)

    @classmethod
    def com_for_translate(cls, children, *args, **kwargs):
        result = cls.com_for_children(children)
        if kwargs.get('clone'):
            return (result + result.translate(*args[0])) / 2
        return result.translate(*args[0])

    @classmethod
    def com_for_mirror(cls, children, *args, **kwargs):
        result = cls.com_for_children(children)
        if kwargs.get('clone'):
            return (result + result.mirror(*args[0])) / 2
        return result.mirror(*args[0])

    @classmethod
    def com_for_rotate(cls, children, *args, **kwargs):
        result = cls.com_for_children(children)
        if kwargs.get('clone'):
            return (result + result.rotate(*args[0])) / 2
        return result.rotate(*args[0])

    @classmethod
    def com_for_scale(cls, children, *args, **kwargs):
        result = cls.com_for_children(children)
        if kwargs.get('clone'):
            return (result + result.scale(*args[0])) / 2
        return result.scale(*args[0])

    @classmethod
    def com_for_union(cls, children, *args, **kwargs):
        return cls.com_for_children(children)

    @classmethod
    def com_for_hull(cls, children, *args, **kwargs):
        return cls.com_for_children(children)

    def translate(self, x, y, z):
        return Vector(self.x + x, self.y + y, self.z + z)

    def scale(self, x, y, z):
        return Vector(self.x * x, self.y * y, self.z * z)

    def mirror(self, x, y, z):
        return Vector(
            self.x * (-1 if x else 1),
            self.y * (-1 if y else 1),
            self.z * (-1 if z else 1),
        )

    def mx(self):
        return self.mirror(1, 0, 0)

    def my(self):
        return self.mirror(0, 1, 0)

    def mz(self):
        return self.mirror(0, 0, 1)

    def rotate(self, ax=0, ay=0, az=0):
        a_sin = sin(ax * pi / 180)
        a_cos = cos(ax * pi / 180)
        x1 = self.x
        y1 = self.y * a_cos - self.z * a_sin
        z1 = self.y * a_sin + self.z * a_cos

        a_sin = sin(ay * pi / 180)
        a_cos = cos(ay * pi / 180)
        x2 = x1 * a_cos + z1 * a_sin
        y2 = y1
        z2 = -x1 * a_sin + z1 * a_cos

        a_sin = sin(az * pi / 180)
        a_cos = cos(az * pi / 180)
        x3 = x2 * a_cos - y2 * a_sin
        y3 = x2 * a_sin + y2 * a_cos
        z3 = z2
        return Vector(x3, y3, z3)

    def rx(self, ax):
        return self.rotate(ax=ax)

    def ry(self, ay):
        return self.rotate(ay=ay)

    def rz(self, az):
        return self.rotate(az=az)

    @lazy
    def norm(self):
        return sqrt(self.x**2 + self.y**2 + self.z**2)

    @lazy
    def normed(self):
        if self.norm == 0:
            return Vector(0, 0, 0)
        return Vector(self.x / self.norm, self.y / self.norm, self.z / self.norm)

    @lazy
    def normal(self):
        return Vector(-self.y, self.x, self.z)

    @lazy
    def as_array(self):
        return [self.x, self.y, self.z]

    @lazy
    def as_array_2d(self):
        return [self.x, self.y]

    @lazy
    def alpha(self):
        return atan2(self.x, self.y)

    def t(self, x=0, y=0, z=0):
        return self.translate(x, y, z)

    def tx(self, x):
        return self.translate(x, 0, 0)

    def ty(self, y):
        return self.translate(0, y, 0)

    def tz(self, z):
        return self.translate(0, 0, z)

    def dot(a, b):
        return a.x * b.x + a.y * b.y + a.z * b.z

    def cross(a, b):
        return Vector(
            a.y * b.z - a.z * b.y,
            a.z * b.x - a.x * b.z,
            a.x * b.y - a.y * b.x,
        )

    def projection(self, pl, pr):
        v1 = pr - pl
        v2 = self - pl

        if v2.norm == 0:
            return self

        new_length = v1.dot(v2) / v1.norm
        result = Vector(
            pl.x + new_length * v1.x / v1.norm,
            pl.y + new_length * v1.y / v1.norm
        )
        return result

    def __sub__(a, b):
        return Vector(a.x - b.x, a.y - b.y, a.z - b.z)

    def __add__(a, b):
        return Vector(a.x + b.x, a.y + b.y, a.z + b.z)

    def __mul__(self, scale):
        return Vector(self.x * scale, self.y * scale, self.z * scale)

    def __truediv__(self, scale):
        return Vector(self.x / scale, self.y / scale, self.z / scale)

    def __floordiv__(self, scale):
        return Vector(self.x // scale, self.y // scale, self.z // scale)

    def __str__(self):
        return "Vector(%f, %f, %f)" % (self.x, self.y, self.z)



# coding: utf8

from math import acos, pi, sqrt

from .base import Node
from .vector import Vector


def point_orientation(a, b, c):
    """Returns the orientation of the triangle a, b, c.

    Return True if a,b,c are oriented clock-wise.
    """
    return (b.x - a.x) * (c.y - a.y) - (c.x - a.x) * (b.y - a.y) >= 0


def line_intersection(a, b, c, d):
    a1 = b.y - a.y
    b1 = a.x - b.x
    c1 = a1 * a.x + b1 * a.y

    a2 = d.y - c.y
    b2 = c.x - d.x
    c2 = a2 * c.x + b2 * c.y

    dt = a1 * b2 - a2 * b1
    if dt == 0:
        return (c + d) / 2
    x = (b2 * c1 - b1 * c2) / dt
    y = (a1 * c2 - a2 * c1) / dt
    return Vector(x, y)


class Path(object):

    def __init__(self, points):
        self.points = []
        for p in points:
            if isinstance(p, list):
                p = Vector(*p)
            assert isinstance(p, Vector), 'Point should be vector or [x, y, z] list'
            self.points.append(p)

    @classmethod
    def _sort_ccw(cls, points):
        return points

    def polygon(self, convexity=2):
        return Node('polygon', None, [p.as_array_2d for p in self.points], convexity=convexity)

    def _triples(self):
        points = self._sort_ccw(self.points)
        total = len(points)
        for i in range(total):
            yield points[i], points[(i + 1) % total], points[(i + 2) % total]

    def offset(self, r=0, fn=16):
        # TODO evaluate fn wisely
        result = []
        for p, c, n in self._triples():
            print(p, c, n)
            is_convex = point_orientation(p, c, n)

            if is_convex == (r < 0):
                b = ((p - c).normed + (n - c).normed).normed
                bcos = (n - c).normed.dot(b)
                bsin = sqrt(1 - bcos ** 2)
                b = b * r / bsin
                result.append(b + c)
            else:
                unit_normal = (c - p).normal.normed
                unit_normal2 = (n - c).normal.normed

                n_prime = c - unit_normal * r
                n_prime2 = c - unit_normal2 * r

                angle = acos((n_prime - c).normed.dot((n_prime2 - c).normed))
                for i in range(fn + 1):
                    alpha = float(i) / fn
                    result.append(
                        (n_prime - c).rz(alpha * angle * 180 / pi) + c
                    )
        return Path(result)

    def round_corners(self, r=0, fn=16):
        # TODO evaluate fn wisely
        if r <= 0:
            return self
        result = []

        for p, c, n in self._triples():
            p_normal = (c - p).normal.normed * r
            n_normal = (n - c).normal.normed * r

            bisect = ((p - c).normed + (n - c).normed).normed
            bcos = (n - c).normed.dot(bisect)
            bsin = sqrt(1 - bcos ** 2)
            bisect = bisect * r / bsin

            center = c + bisect

            is_convex = point_orientation(p, c, n)

            angle = acos(p_normal.normed.dot(n_normal.normed))
            # result.append(center)
            for i in range(fn + 1):
                alpha = float(i) / fn
                if is_convex:
                    result.append(center - p_normal.rz(alpha * angle * 180 / pi))
                else:
                    result.append(center + p_normal.rz(-alpha * angle * 180 / pi))
        return Path(result)

    def translate(self, x=0, y=0, z=0):
        return Path([p.translate(x, y, z) for p in self.points])

    def tx(self, x):
        return self.translate(x=x)

    def ty(self, y):
        return self.translate(y=y)

    def tz(self, z):
        return self.translate(z=z)

    def extrude(self, *args, **kwargs):
        return self.polygon().extrude(*args, **kwargs)


def stitch_slices(slices, convexity=10):
    slice_count = len(slices)
    points_per_slice = len(slices[0].points)
    z0 = slices[0].points[0].z
    z1 = slices[-1].points[0].z

    points = []
    faces = []
    # TODO add assertion for all data (flat, z increasing)
    for s in slices:
        assert len(s.points) == points_per_slice, 'Slices should be same size'
        for p in s.points:
            points.append(p.as_array)

    points.append([0, 0, z0])
    points.append([0, 0, z1])

    z0idx = len(points) - 2
    z1idx = len(points) - 1

    for i in range(points_per_slice):
        faces.append([z0idx, (i + 1) % points_per_slice, i])
        faces.append([
            z1idx,
            points_per_slice * (slice_count - 1) + ((i - 1) % points_per_slice),
            points_per_slice * (slice_count - 1) + i,
        ])

    for layer in range(slice_count - 1):
        for i in range(points_per_slice):
            faces.append([
                points_per_slice * layer + i,
                points_per_slice * layer + (i + 1) % points_per_slice,
                points_per_slice * (layer + 1) + i,
            ])
            faces.append([
                points_per_slice * (layer + 1) + i,
                points_per_slice * layer + (i + 1) % points_per_slice,
                points_per_slice * (layer + 1) + (i + 1) % points_per_slice,
            ])

    return Node('polyhedron', None, points=points, faces=faces, convexity=convexity)


def blend_polar(a, b, alpha):
    points = []
    for p in a.points:
        points.append((p.alpha, p, None))
    for p in b.points:
        points.append((p.alpha, None, p))
    points.sort(key=lambda x: x[0])

    result = []

    def _nearest(points, i, idx):
        j = i
        pl = None
        while pl is None:
            pl = points[j][idx]
            j = (j - 1) % len(points)

        j = i
        pr = None
        while pr is None:
            pr = points[j][idx]
            j = (j + 1) % len(points)
        return pl, pr

    def _intersection(a, b, c, d):
        a1 = b.y - a.y
        b1 = a.x - b.x
        c1 = a1 * a.x + b1 * a.y

        a2 = d.y - c.y
        b2 = c.x - d.x
        c2 = a2 * c.x + b2 * c.y

        dt = a1 * b2 - a2 * b1
        if dt == 0:
            return (c + d) / 2
        x = (b2 * c1 - b1 * c2) / dt
        y = (a1 * c2 - a2 * c1) / dt
        return Vector(x, y)

    result = []
    for i, (angle, p1, p2) in enumerate(points):
        if p1 is not None:
            pl, pr = _nearest(points, i, 2)
            prj = _intersection(Vector(), p1 * 10000, pl, pr)
            x = p1.x * (1.0 - alpha) + prj.x * alpha
            y = p1.y * (1.0 - alpha) + prj.y * alpha
            z = p1.z * (1.0 - alpha) + (pl.z + pr.z) * 0.5 * alpha
        elif p2 is not None:
            pl, pr = _nearest(points, i, 1)
            prj = _intersection(Vector(), p2 * 10000, pl, pr)
            x = p2.x * alpha + prj.x * (1.0 - alpha)
            y = p2.y * alpha + prj.y * (1.0 - alpha)
            z = p2.z * alpha + (pl.z + pr.z) * 0.5 * (1.0 - alpha)

        result.append(Vector(x, y, z))

    return Path(result)


def extrude_blend_polar(path1, path2, h=1, fn=16):
    slices = []
    for i in range(fn + 1):
        alpha = float(i) / fn
        z = h * alpha
        slices.append(blend_polar(path1, path2, alpha).tz(z))
    return stitch_slices(slices)



# coding: utf-8
from math import pi, tan, cos, sin
from .local_logging import get_logger
from .scad import cylinder, polyhedron

logger = get_logger(__name__)

# nuts = lambda: None  # noqa
#
# nuts.m4.set(
#     height=3.2,
#     width=7.,
#     diameter=8.,
#     internal_diameter=3.95,
#     screw_diameter=4.1,
# )


def deg2rad(deg):
    return deg * pi / 180


def rounded_box(x, y, z, r=5, fn=32):
    result = cylinder(r=r, h=z, fn=fn).t(r, r, 0).mx(x / 2, clone=True).my(y / 2, clone=True).hull()
    result.com.x = x / 2
    result.com.y = y / 2
    result.com.z = z / 2
    result.size.x = x
    result.size.y = y
    result.size.z = z
    return result


def _thread(length, pitch, diameter, fn=32, tolerance=0, theta=60):
    segments_per_revolution = fn

    if length < pitch / 2:
        raise Exception('Length is too small')

    # internal stuff starts here
    D = diameter
    P = pitch
    H = P / (2 * tan(theta / 2 * pi / 180))

    # TODO change radiuses to match right dimension
    r = (D - H * 5 / 8) / 2
    R = D / 2

    # shift for internal helix
    s = P / 4
    S = P / 8

    # Four points for helixes
    h1 = []
    h2 = []
    H1 = []
    H2 = []

    L = length + P * 2
    revolutions = L / P
    segments = int(revolutions * segments_per_revolution)
    for segment in range(segments + 1):
        angle = segment / segments_per_revolution * pi * 2
        height = segment / segments * L - P

        x, y = (r + tolerance) * cos(angle), (r + tolerance) * sin(angle)
        X, Y = (R + tolerance) * cos(angle), (R + tolerance) * sin(angle)

        h1.append([x, y, height - s / 2])
        h2.append([x, y, height + s / 2])

        H1.append([X, Y, height - S / 2 + P / 2])
        H2.append([X, Y, height + S / 2 + P / 2])

    points = h1 + h2 + H1 + H2
    points.append([0, 0, -P])
    points.append([0, 0, length + P])

    faces = []
    spr = segments_per_revolution
    stot = segments + 1
    for i in range(stot - 1):
        for stot_factor in [0, 1, 2]:
            # current index
            ii = i + stot * stot_factor

            # internal helix line
            faces.append([ii, ii + 1, ii + stot])
            faces.append([ii + 1, ii + stot + 1, ii + stot])

        # external to next internal helix
        if i < stot - spr - 1:
            ii = i + stot * 3
            faces.append([ii, ii + 1, i + spr])
            faces.append([ii + 1, i + spr + 1, i + spr])

    bottom_idx = len(points) - 2
    for i in range(spr):
        faces.append([i + 1, i, bottom_idx])
    for i in range(3):
        faces.append([bottom_idx, i * stot, stot * (i + 1)])
    faces.append([bottom_idx, stot * 3, spr])

    top_idx = len(points) - 1
    for i in range(spr):
        faces.append([stot * 4 - 1 - i - 1, stot * 4 - 1 - i, top_idx])

    lst_idx = stot * 4 - 1
    for i in range(3):
        faces.append([top_idx, lst_idx - stot * i, lst_idx - stot * (i + 1)])
    faces.append([top_idx, lst_idx - stot * 3, lst_idx - spr])

    faces = [reversed(f) for f in faces]
    result = polyhedron(
        points=points,
        faces=faces,
        convexity=revolutions,
    )

    result = result.intersection(cylinder(d=(R + abs(tolerance * 2)) * 2, h=length))
    result.pitch = pitch
    result.D = R * 2
    result.d = r * 2
    return result


def _diameter_to_pitch(diameter):
    # TODO faster search
    table = [
        [0, 0.25],
        [1.4, 0.3],
        [1.6, 0.35],
        [2, 0.4],
        [2.2, 0.45],
        [3, 0.5],
        [3.5, 0.6],
        [4, 0.7],
        [5, 0.8],
        [6, 1],
        [8, 1.25],
        [10, 1.5],
        [12, 1.75],
        [14, 2],
        [18, 2.5],
        [24, 3],
        [30, 3.5],
        [36, 4],
        [42, 4.5],
        [48, 5],
        [56, 5.5],
        [64, 6],
        [10000, 6],
    ]
    prev_pitch = 0.25
    for d, pitch in table:
        if diameter == d:
            return pitch
        if diameter < d:
            return prev_pitch
        prev_pitch = pitch


def _pitch_to_tolerance(pitch):
    return pitch * 0.0866 * 7 / 8 * 0.5 * 0.5  # 2.5% of H


def m_thread_external(d=4, h=10, fn=64):
    pitch = _diameter_to_pitch(d)
    tolerance = -_pitch_to_tolerance(pitch)
    result = _thread(h, pitch, d, tolerance=tolerance, fn=fn)
    return result


def m_thread_internal_hole(d=4, h=10, fn=64):
    pitch = _diameter_to_pitch(d)
    tolerance = _pitch_to_tolerance(pitch)
    result = _thread(h, pitch, d, tolerance=tolerance, fn=fn)
    return result
